#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import

import sys

from pip._internal.utils import misc


def get_prog():
    return "pipm"


# patch for program name
misc.get_prog = get_prog

from pip._internal import commands

from pipm.commands import (
    InstallCommandPlus,
    UninstallCommandPlus,
    UpdateCommand,
    FreezeCommandPlus,
)


def _update_command_info(cls, *names):
    old_info = commands.commands_dict.get(names[0])
    for n in names:
        commands.commands_dict[n] = commands.CommandInfo(
            cls.__module__, cls.__name__, old_info.summary if old_info else cls.summary
        )


# replace commands
for args in [
    (InstallCommandPlus, "install", "i", "add"),
    (UninstallCommandPlus, "uninstall", "rm"),
    (FreezeCommandPlus, "freeze", "save"),
    (UpdateCommand, "update", "upgrade", "u"),
]:
    _update_command_info(*args)


def is_venv():
    return hasattr(sys, "real_prefix") or (
        hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix
    )


from pip._internal.cli.main import main as pip_main


def main():
    if not is_venv():
        raise Exception("Please install `pipm` inside virtualenv")
    pip_main()


if __name__ == "__main__":
    sys.exit(main())



import errno
import os


def _new_line(filename):
    """
        append `\n` to the end of the file if it doesn't exist
    Args:
        filename:
    """
    with open(filename, "ab+") as f:
        # check for empty file
        f.seek(0)
        if f.read(1) == b"" and f.read() == b"":
            return
        try:
            f.seek(-1, os.SEEK_END)
            if f.read(1) != b"\n":
                f.write(b"\n")
        except OSError:
            pass


def get_patterns(*envs):
    """

    Args:
        *envs:

    Returns:
        list:
    """
    patterns = []
    for env in envs:
        patterns.append("{}requirements.txt".format("" if env == "base" else env + "-"))
        patterns.append(os.path.join("requirements", "{}.txt".format(env)))
        if env != "base":
            patterns.append("requirements-{}.txt".format(env))
    return patterns


def parse_comes_from(comes_from, env):
    """
        parse comesfrom if valid otherwise return the filename for the environment
    Args:
        comes_from ([str]):
        env (str):

    Returns:
        str, int: filename and line number
    """
    if comes_from:
        comes_from = comes_from.split()
        return comes_from[1], int(comes_from[3].strip(")"))
    filename = get_req_filename(env)
    return filename, None


def get_req_filename(env=""):
    """

    Args:
        env:

    Returns:
        str:
    """
    BASE_PTRN = ("base",)
    DEV_PTRN = ("dev", "development")
    TEST_PTRN = ("test",)
    PROD_PTRN = ("prod", "production")

    envs = (
        DEV_PTRN
        if env == "dev"
        else PROD_PTRN
        if env == "prod"
        else TEST_PTRN
        if env == "test"
        else (env,)
        if env
        else BASE_PTRN
    )
    paths = get_patterns(*envs)

    fname = get_env_reqfile(
        *paths,
        base_file_name="" if not env else get_env_reqfile(*get_patterns(*BASE_PTRN))
    )
    _new_line(fname)
    return fname


def get_env_reqfile(*paths, **kw):
    """
        from the list of paths return the one that exists. If it doesn't exists then create with appropriate
        starter line
    Args:
        env:
        base_file_name:
        *paths:

    Returns:
        str:
    """
    base_file_name = kw.get("base_file_name", "")
    requirements_dir = os.path.join("requirements", "")
    for path in paths:
        if os.path.exists(path):
            if path == "requirements.txt":
                base_path = os.path.join("requirements", "base.txt")
                if os.path.exists(base_path):
                    return base_path
            return path

    # create file if it doesnt exist
    filename = paths[0]  # prefer the first pattern in the list

    # if requirements directory exists then prefer creating files inside that one
    if os.path.exists(os.path.join(os.curdir, "requirements")):
        for path in paths:
            if requirements_dir in path:
                filename = path

    if os.path.dirname(filename) and not os.path.exists(os.path.dirname(filename)):
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError as exc:  # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise

    if not os.path.exists(filename):
        with open(filename, "wb") as f:
            if base_file_name:
                if filename != base_file_name:
                    f.write(
                        "-r {}".format(
                            "base.txt"
                            if requirements_dir in filename
                            else base_file_name
                        ).encode("utf-8")
                    )

    return filename



from __future__ import absolute_import, unicode_literals

import logging
import os
from collections import OrderedDict

from pip._internal.network.session import PipSession
from pip._internal.req import req_file
from pip._internal.req.constructors import install_req_from_parsed_requirement
from pip._internal.req.req_file import get_file_content
from pip._internal.req.req_install import InstallRequirement

from . import operations, setup_cfg
from .classes import FileRequirement
from .file_utils import get_req_filename

try:
    from typing import List, Dict
except ImportError:
    pass

logger = logging.getLogger(__name__)


def get_req_filenames():
    """return all requirement files in the current project that matches the standard requirements filename pattern"""
    filenames = set()

    # if requirements directory exists then add those
    req_dir = os.path.join(os.curdir, "requirements")
    if os.path.exists(req_dir):
        for fn in os.listdir(req_dir):
            filename = os.path.join("requirements", fn)
            if os.path.isfile(filename):
                if filename.endswith(".txt"):
                    filenames.add(filename)
    else:
        # walk current directory
        for filename in os.listdir(os.curdir):
            if os.path.isfile(filename):
                if filename.endswith("requirements.txt"):
                    filenames.add(filename)
                elif filename.startswith("requirements-"):
                    filenames.add(filename)

    return filenames


def _uniq_resources(reqs):
    # type: (List[InstallRequirement]) -> Dict[str, InstallRequirement]
    uniq_reqs = OrderedDict()
    for req in reqs:
        if req.name in uniq_reqs:  # req.name = "xdis"
            old_req = uniq_reqs[req.name]
            if (
                not req.comes_from and old_req.comes_from
            ):  # req.comes_from = '-r requirements.txt (line 1)'
                req.comes_from = old_req.comes_from
        uniq_reqs[req.name] = req
    return uniq_reqs


def cluster_to_file_reqs(reqs, env):
    # type: (Dict[str, InstallRequirement], str) -> Dict[str, List[FileRequirement]]
    filereqs = OrderedDict()
    for req in reqs.values():
        freq = FileRequirement(req, env)
        if freq.filename not in filereqs:  # default
            filereqs[freq.filename] = []
        filereqs[freq.filename].append(freq)

    for filename in get_req_filenames():
        if not filereqs.get(filename):
            filereqs[filename] = []

    return filereqs


def get_parsed_requirements():
    # type: () -> (PipSession, Dict[str, InstallRequirement])
    session = PipSession()
    reqs = []
    for file in get_req_filenames():
        reqs += list(req_file.parse_requirements(file, session=session))

    inst_reqs = [install_req_from_parsed_requirement(req) for req in reqs]
    return session, _uniq_resources(inst_reqs)


def save(env="", user_reqs=None, uninstall=False):
    # type: (str, List[InstallRequirement], bool) -> None
    """
        save installed requirements which is missing in the requirements files
    Args:
        user_reqs: list of strings that are explicitly given as argument to the user installing
    """
    if user_reqs:
        setup_cfg.add_requirements(user_reqs, env)
    if uninstall:
        setup_cfg.remove_requirements()

    write_to_req_files(env)


def write_to_req_files(env):
    # create base file if it doesnt exists
    env_filename = get_req_filename(env)
    session, reqs = get_parsed_requirements()
    file_reqs = cluster_to_file_reqs(reqs, env)

    installations = operations.get_frozen_reqs()

    # first step process the requirements and split them into separate for each of the file
    for filename in file_reqs:  # type: str
        _, content = get_file_content(filename, session=session)

        orig_lines = enumerate(content.splitlines(), start=1)
        joined_lines = req_file.join_lines(orig_lines)
        lines = OrderedDict(joined_lines)

        # 1. save new requirements
        if filename == env_filename:
            installed = set(installations.keys()).difference(set(reqs.keys()))
            for new_req in installed:
                line_num = len(lines) + 1
                lines[line_num] = str(installations[new_req]).strip()

        for req in file_reqs[filename]:
            frozenrequirement = installations.get(req.req.name)
            if frozenrequirement:
                # 2. updates
                lines[req.line_num] = str(frozenrequirement).strip()
            else:
                # 3. removals
                lines.pop(req.line_num)

        # 4. finally write to file
        with open(filename, "wb") as f:
            for line in lines:
                cnt = lines[line].strip()
                if cnt:
                    f.write((cnt + "\n").encode("utf-8"))



from __future__ import absolute_import
import threading

from pip._internal.cli.cmdoptions import make_target_python
from pip._internal.commands.uninstall import UninstallCommand
from pip._internal.commands.freeze import FreezeCommand
from pip._internal.commands import install

from pipm.operations import get_orphaned_packages
from . import file, file_utils

INTERACTIVE_UPDATE = "--interactive-update"

THREAD_GLOB = threading.local()


def store_req_environment(_, opt_str, __, parser, *___, **____):
    parser.values.req_environment = opt_str.strip("-")


orig_install_given_reqs = install.install_given_reqs


def patched_install_given_reqs(to_install, *args, **kwargs):
    from pip._internal.utils.logging import indent_log

    accepted_reqs = []
    with indent_log():
        confirm_update = getattr(THREAD_GLOB, "interactive_update", False)
        for requirement in to_install:
            if confirm_update:
                want_to_install = input(
                    "Do you want to update {}? [Y/n]".format(requirement)
                )
                if str(want_to_install).lower() in {"no", "n"}:
                    continue
            accepted_reqs.append(requirement)
    return orig_install_given_reqs(accepted_reqs, *args, **kwargs)


# patch the original function
install.install_given_reqs = patched_install_given_reqs


class InstallCommandPlus(install.InstallCommand):
    def __init__(self, *args, **kw):
        """

        Args:
            *args:
            **kw:
        """
        super(InstallCommandPlus, self).__init__(*args, **kw)

        cmd_opts = self.cmd_opts

        for env, help in (
            ("all", "requirements from all environments."),
            ("dev", "work in development environment"),
            ("test", "work in testing environment"),
            ("prod", "work in production environment"),
        ):
            cmd_opts.add_option(
                "--{}".format(env),
                dest="req_environment",
                action="callback",
                callback=store_req_environment,
                help=help,
            )

        cmd_opts.add_option(
            "--env",
            dest="req_environment",
            action="store",
            help="add as seperate set of dependency other than dev/test/prod environments",
        )

    def update_opts_args(self, options, args):
        env = options.req_environment
        upgrade = options.upgrade

        if env == "all":
            req_args = []
            for req in file.get_req_filenames():
                req_args.append("-r")
                req_args.append(req)
        else:
            req_args = ["-r", file_utils.get_req_filename(env)]
        options, args = super(InstallCommandPlus, self).parse_args(req_args)
        options.req_environment = env
        options.upgrade = upgrade
        options.no_save = True
        return options, args

    def parse_args(self, args):
        """
            when no argument given it fills with `-r requirements.txt` as default
        Args:
            args (list):

        Returns:
            options, list:
        """
        options, args = super(InstallCommandPlus, self).parse_args(args)
        if not options.requirements and (
            (len(args) == 1 and set(args) == {"--all"}) or not args
        ):
            options, args = self.update_opts_args(options, args)
        return options, args

    def run(self, options, args):
        """
            wrapper around pip.commands.InstallCommand
        Args:
            options:
            args:

        Returns:
            pip.req.RequirementSet:
        """
        result = super(InstallCommandPlus, self).run(options, args)
        if not hasattr(options, "no_save"):
            session = self.get_default_session(options)
            target_python = make_target_python(options)
            finder = self._build_package_finder(
                options,
                session,
                target_python,
                ignore_requires_python=options.ignore_requires_python,
            )
            user_reqs = self.get_requirements(
                args,
                options,
                finder,
                session,
                check_supported_wheels=not options.target_dir,
            )
            # save changes to file if any
            file.save(env=options.req_environment, user_reqs=user_reqs)

        return result


class FreezeCommandPlus(FreezeCommand):
    """
        A wrapper around standard freeze command that updates currently installed packages to requirement files
        after showing the packages list in standard output.
    """

    def run(self, options, args):
        res = super(FreezeCommandPlus, self).run(options, args)

        file.save()

        return res


class UninstallCommandPlus(UninstallCommand):
    """
        a drop-in replacement of pip's uninstall command. It removes the entry from requirements file after a package
        is removed.
    """

    def run(self, options, args):
        removable_pkgs = get_orphaned_packages(args)

        if removable_pkgs:
            print(
                "Following packages are no longer required by any of the installed packages: \n",
                "\n".join(removable_pkgs),
            )

        res = super(UninstallCommandPlus, self).run(
            options, (args + list(removable_pkgs))
        )

        file.save(uninstall=True)

        return res


class UpdateCommand(InstallCommandPlus):
    usage = """
          %prog [environment-to-update]
          %prog [package-names-to-update]"""

    summary = "Update packages (equivalent to that of `install` with --upgrade)"

    def __init__(self, *args, **kw):
        """

        Args:
            *args:
            **kw:
        """
        super(UpdateCommand, self).__init__(*args, **kw)

        cmd_opts = self.cmd_opts

        cmd_opts.add_option(
            "--auto-update",
            dest="interactive_update",
            default=True,
            action="store_false",
            help="Update packages without user input",
        )

        cmd_opts.add_option(
            "--latest",
            dest="update_to_latest",
            default=False,
            action="store_true",
            help="Update packages to latest version instead of aligning with setup.cfg",
        )

    def update_opts_args(self, options, args):
        env = options.req_environment
        upgrade = options.upgrade
        to_latest = options.update_to_latest

        from pipm.setup_cfg import get_requirements

        reqs = get_requirements(env)

        req_args = list(reqs.keys() if to_latest else reqs.values())
        options, args = super(InstallCommandPlus, self).parse_args(
            req_args
        )  # call the parent's method and not recursion
        options.req_environment = env
        options.upgrade = upgrade
        return options, args

    def parse_args(self, args):
        """
            when no argument given it fills with `-r requirements.txt` as default
        Args:
            args (list):

        Returns:
            options, list:
        """
        options, args = super(UpdateCommand, self).parse_args(args + ["--upgrade"])
        THREAD_GLOB.interactive_update = options.interactive_update

        return options, args



from pip._internal.req import InstallRequirement

from .file_utils import parse_comes_from


class FileRequirement:
    def __init__(self, req, env):
        # type: (InstallRequirement, str) -> None
        filename, line_num = parse_comes_from(req.comes_from, env)
        self.req = req
        self.filename = filename
        self.line_num = line_num

    def __repr__(self):
        return "{}".format(self.req)



import codecs
import os

from setuptools import setup, find_packages
import pipm

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    return codecs.open(os.path.join(here, *parts), "r").read()


setup(
    name="pipm",
    version=pipm.__version__,
    description="Wrapper around pip commands to auto-update requirements file",
    long_description="Checkout the README https://github.com/jnoortheen/pipm",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Build Tools",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    keywords="easy_install distutils setuptools egg virtualenv requirements",
    author="noortheen",
    author_email="jnoortheen@gmail.com",
    url="https://github.com/jnoortheen/pipm",
    homepage="https://github.com/jnoortheen/pipm",
    license="MIT",
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    entry_points={"console_scripts": ["pipm=pipm.__main__:main"]},
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*",
    # tests_require=tests_require,
    install_requires=["pip>=20.0", "six", 'typing;python_version<"3.6"']
    # extras_require={
    #     'testing': tests_require,
    # },
)
